import os
import sys

from contextlib import ExitStack, contextmanager
from doctest import ELLIPSIS, REPORT_NDIFF, NORMALIZE_WHITESPACE
from sybil import Sybil
from sybil.parsers.codeblock import CodeBlockParser
from sybil.parsers.doctest import DocTestParser
from tempfile import TemporaryDirectory
from types import ModuleType

import pytest

DOCTEST_FLAGS = ELLIPSIS | NORMALIZE_WHITESPACE | REPORT_NDIFF


@contextmanager
def syspath(directory):
    try:
        sys.path.insert(0, directory)
        yield
    finally:
        assert sys.path[0] == directory
        del sys.path[0]


@contextmanager
def sysmodules():
    modules = sys.modules.copy()
    try:
        yield
    finally:
        sys.modules = modules


class ExampleModule:
    def __init__(self, path):
        self.path = path

    def __call__(self, contents):
        with open(self.path, 'w', encoding='utf-8') as fp:
            fp.write(contents)


@pytest.fixture
def example():
    with ExitStack() as resources:
        tmpdir = resources.enter_context(TemporaryDirectory())
        resources.enter_context(sysmodules())
        resources.enter_context(syspath(tmpdir))
        path = os.path.join(tmpdir, 'example.py')
        yield ExampleModule(path)


class DoctestNamespace:
    def setup(self, namespace):
        # The doctests in .rst files require that they mimic being executed in
        # a particular module.  The stdlib doctest functionality creates its
        # own globals namespace, unattached to any specific module object.
        # This causes coordination problems between the apparent globals that
        # the doctest sees, and public()'s implementation.
        #
        # We can't make them the same namespace because doing so violates
        # other assumptions in the public() function's code, but we can set
        # things up to be close enough for the doctest to pass.
        #
        # We use two techniques to make this work.  First, we create a test
        # module and ensure that its string name is assigned to the
        # namespace's __name__ attribute.  We also ensure that the module by
        # that name is in the sys.modules cache (and cleaned up in the
        # teardown).
        #
        # The second thing we need to do is to ensure that the module and the
        # namespace the doctest is executed in, share the same list object in
        # their __all__ attribute.  Now, generally public() will create
        # __all__ if it doesn't exist, but we can test that in the unittests,
        # so it's good enough to just initialize both name bindings to the
        # same list object here.
        #
        # There is some further discussion in this Sybil ticket:
        # https://github.com/cjw296/sybil/issues/21
        self._testmod = ModuleType('testmod')
        namespace['__name__'] = self._testmod.__name__
        sys.modules[self._testmod.__name__] = self._testmod
        self._testmod.__all__ = namespace['__all__'] = []

    def teardown(self, namespace):
        del sys.modules[self._testmod.__name__]


namespace = DoctestNamespace()


pytest_collect_file = Sybil(
    parsers=[
        DocTestParser(optionflags=DOCTEST_FLAGS),
        CodeBlockParser(),
        ],
    pattern='*.rst',
    setup=namespace.setup,
    ).pytest()



import sys

from typing import Any, Optional

from .types import ModuleAware


# http://bugs.python.org/issue26632
def public(thing: Optional[ModuleAware] = None,
           **kws: Any
           ) -> Optional[ModuleAware]:
    """Add a name or names to __all__

    There are two forms of use for this function.  Most commonly it will
    be used as a decorator on a class or function at module scope.  In
    this case, `thing` will be an object with both __module__ and
    __name__ attributes, and the name is added to the module's __all__
    list, creating that if necessary.

    When used in its function call form, `thing` will be None.  __all__
    is looked up in the globals at the function's call site, and each
    key in the keyword arguments is added to the __all__.  In addition,
    the key will be bound to the value in the globals.

    Only one or the other format may be used.

    :param thing: None, or an object with both a __module__ and a __name__
        argument.
    :param kws: Keyword arguments.
    :return: The original `thing` object.
    :raises ValueError: When the inputs are invalid, or this function finds
        a non-list __all__ attribute.
    """
    # 2020-07-14(warsaw): I considered using inspect.getmodule() here but
    # looking at its implementation, I feel like it does a ton of unnecessary
    # work in the oddball cases (i.e. where the object does not have an
    # __module__ attribute).  Because @public runs at module import time, and
    # because I'm not really sure we even want to support those oddball cases,
    # I'm taking the more straightforward approach of just looking the module
    # up in sys.modules.  That should be good enough for our purposes.
    mdict = (
        # The function call syntax.
        sys._getframe(1).f_globals
        if thing is None
        # The decorator syntax.
        else sys.modules[thing.__module__].__dict__)
    dunder_all = mdict.setdefault('__all__', [])
    if not isinstance(dunder_all, list):
        raise ValueError(f'__all__ must be a list not: {type(dunder_all)}')
    if thing is None:
        # The function call form.
        for key, value in kws.items():
            # This overwrites any previous similarly named __all__ entry.
            if key not in dunder_all:
                dunder_all.append(key)
            # We currently do not check for duplications in the globals.
            mdict[key] = value
    else:
        # I think it's impossible to use the @public decorator and pass in
        # keyword arguments.  Not quite syntactically impossible, but you'll
        # get a TypeError if you try it, before you even get to this code.
        assert len(kws) == 0, (
            'Keyword arguments are incompatible with use as decorator')
        if thing.__name__ not in dunder_all:
            dunder_all.append(thing.__name__)
    return thing



import sys

from .types import ModuleAware


def private(thing: ModuleAware) -> ModuleAware:
    """Remove names from __all__

    This decorator documents private names and ensures that the names do not
    appear in the module's __all__.

    :param thing: An object with both a __module__ and a __name__ argument.
    :return: The original `thing` object.
    :raises ValueError: When this function finds a non-list __all__ attribute.
    """
    mdict = sys.modules[thing.__module__].__dict__
    dunder_all = mdict.setdefault('__all__', [])
    if not isinstance(dunder_all, list):
        raise ValueError(f'__all__ must be a list not: {type(dunder_all)}')
    if thing.__name__ in dunder_all:
        dunder_all.remove(thing.__name__)
    return thing



"""public -- @public for populating __all__"""

from setup_helpers import get_version, require_python
from setuptools import setup, find_packages

require_python(0x30600f0)
__version__ = get_version('public/__init__.py')


with open('README.rst') as fp:
    readme = fp.read()


setup(
    name='atpublic',
    version=__version__,
    author='Barry Warsaw',
    author_email='barry@python.org',
    description=__doc__,
    long_description=readme,
    long_description_content_type='text/x-rst',
    license='Apache 2.0',
    keywords='__all__ public',
    url='http://public.readthedocs.io/',
    packages=find_packages(where='.', exclude=['tests*', 'docs']),
    include_package_data=True,
    # readthedocs builds fail unless zip_safe is False.
    zip_safe=False,
    python_requires='>=3.6',
    install_requires=[
        'typing_extensions;python_version<"3.8"',
        ],
    project_urls={
        'Documentation': 'https://public.readthedocs.io',
        'Source': 'https://gitlab.com/warsaw/public.git',
        'Tracker': 'https://gitlab.com/warsaw/public/issues',
        },
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Development Status :: 6 - Mature',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        ],
    )



try:
    from typing import Protocol
except ImportError:                                 # pragma: nocover
    # Python < 3.8
    from typing_extensions import Protocol  # type: ignore


class ModuleAware(Protocol):
    __module__: str
    __name__: str
